#ifndef Basic3DVector_H
#define Basic3DVector_H

#include "Basic2DVector.h"
#include <iosfwd>
#include <cmath>

template < class T> 
class Basic3DVector {

public:
  // default constructor
  Basic3DVector() : theX(0), theY(0), theZ(0){}

  // copy constructor from same type
  Basic3DVector( const Basic3DVector & p) : 
    theX(p.x()), theY(p.y()), theZ(p.z()) {}

  // constructor from 2D vector
  Basic3DVector( const Basic2DVector<T> & p) : 
    theX(p.x()), theY(p.y()), theZ(0) {}

#ifndef CMS_NO_TEMPLATE_MEMBERS
  //  // copy constructor and implicit conversion from other point classes 
  //  template <class OtherPoint> 
  //  Basic3DVector( const OtherPoint& p) : 
  //    theX(p.x()), theY(p.y()), theZ(p.z()) {}
#endif 

  Basic3DVector( const T& x, const T& y, const T& z) : 
    theX(x), theY(y), theZ(z) {}

  T x() const { return theX;}
  T y() const { return theY;}
  T z() const { return theZ;}
  T mag2() const { return theX*theX + theY*theY + theZ*theZ;}
  T r() const    { return sqrt( mag2());}
  T mag() const  { return r();}

  // perp2 and perp for CLHEP interface compatibility
  T perp2() const { return x()*x() + y()*y();}
  T perp() const { return sqrt( perp2());}

  // recommended name 
  T transverse() const { return perp();}

  T phi() const  { return theX == 0.0 && theY == 0.0 ? 0.0 : atan2(theY,theX);}

  T theta() const {
    return x() == 0.0 && y() == 0.0 && z() == 0.0 ? 0.0 : atan2(transverse(),z());
  }

  // pseudorapidity
  // T eta() const { return -log( tan( theta()/2.));} 
  T eta() const { T x(z()/perp()); return log(x+sqrt(x*x+1));} // faster 

  // Unit vector parallel to this.
  Basic3DVector unit() const {
    T my_mag = mag();
    return my_mag == 0 ? *this : Basic3DVector(*this) / my_mag;
  }

  Basic3DVector& operator+= ( const Basic3DVector& p) {
    theX += p.x();
    theY += p.y();
    theZ += p.z();
    return *this;
  } 

  Basic3DVector& operator-= ( const Basic3DVector& p) {
    theX -= p.x();
    theY -= p.y();
    theZ -= p.z();
    return *this;
  } 

  Basic3DVector operator-() const { return Basic3DVector(-x(),-y(),-z());}

  Basic3DVector& operator*= ( const T& t) {
    theX *= t;
    theY *= t;
    theZ *= t;
    return *this;
  } 

  Basic3DVector& operator/= ( const T& t) {
    theX /= t;
    theY /= t;
    theZ /= t;
    return *this;
  } 

  T dot( const Basic3DVector& v) const { 
    return x()*v.x() + y()*v.y() + z()*v.z();
  }

  Basic3DVector cross( const Basic3DVector& v) const {
    return Basic3DVector( y()*v.z() - v.y()*z(), 
			  z()*v.x() - v.z()*x(), 
			  x()*v.y() - v.x()*y());
  }

private:
  T theX;
  T theY;
  T theZ;
  
};

template <class T>
std::ostream & operator<<( std::ostream& s, const Basic3DVector<T>& v) {
  return s << " (" << v.x() << ',' << v.y() << ',' << v.z() << ") ";
} 


template <class T>
Basic3DVector<T> operator+( const Basic3DVector<T>& v1, 
			    const Basic3DVector<T>& v2) {
  return Basic3DVector<T>(v1) += v2;
}

template <class T>
Basic3DVector<T> operator-( const Basic3DVector<T>& v1, 
			    const Basic3DVector<T>& v2) {
  return Basic3DVector<T>(v1) -= v2;
}

template <class T>
T operator*( const Basic3DVector<T>& v1, const Basic3DVector<T>& v2) {
  return v1.dot(v2);
}

template <class T>
Basic3DVector<T> operator*( const Basic3DVector<T>& v, float t) {
  return Basic3DVector<T>(v) *= t;
}

template <class T>
  Basic3DVector<T> operator*( float t, const Basic3DVector<T>& v) {
  return v*t;
}

template <class T>
Basic3DVector<T> operator*( const Basic3DVector<T>& v, double t) {
  return Basic3DVector<T>(v) *= t;
}

template <class T>
Basic3DVector<T> operator*( double t, const Basic3DVector<T>& v) {
  return v*t;
}

template <class T>
Basic3DVector<T> operator/( const Basic3DVector<T>& v, float t) {
  return Basic3DVector<T>(v) /= t;
}

template <class T>
Basic3DVector<T> operator/( const Basic3DVector<T>& v, double t) {
  return Basic3DVector<T>(v) /= t;
}

#endif // Basic3DVector_H


